//////////////////////////////////////////////////////////////////////////// 
// IPD - Image Processing Design Toolbox
//
// Copyright (c) by Dr. Eng. (J) Harald Galda, 2009
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
////////////////////////////////////////////////////////////////////////////


#include "IPD.h"
#include "Filter.h"
#include "Dilation.h"
#include "Erosion.h"


// MorphologicalFilter performs a morphological dilation, erosion, closing, opening, top hat
// or bottom hat filtering. This function is intended to be called by Scilab. Calling syntax
// syntax in Scilab:
// FilteredImage = call('MorphologicalFilter', ...
//                      OriginalImage, ...
//                      1, ...
//                      'd', ...
//                      NumberOfColumns, ...
//                      2, ...
//                      'i', ...
//                      NumberOfRows, ...
//                      3, ...
//                      'i', ...
//                      FilterType, ...
//                      4, ...
//                      'i', ...
//                      StructureElementWidth, ...
//                      5, ...
//                      'i', ...
//                      StructureElementHeight, ...
//                      6, ...
//                      'i', ...
//                      StructureElementData, ...
//                      7, ...
//                      'd', ...
//                      'out', ...
//                      [NumberOfRows, NumberOfColumns], ...
//                      8, ...
//                      'd')'; 
extern "C" __declspec(dllexport)
int MorphologicalFilter(double* OriginalImage, 
                        TUInt* NumberOfColumns, 
                        TUInt* NumberOfRows,
                        TUInt* FilterType,
                        TUInt* StructureElementWidth, 
                        TUInt* StructureElementHeight, 
                        double* StructureElementData,
                        double* FilteredImage)
{
 // Filter is initialized.

 CFilter* FirstPass = NULL;

 CFilter* SecondPass = NULL;

 switch(*FilterType)
 {
  case FILTER_DILATE:
  {
   FirstPass = new CDilation(*StructureElementWidth, 
                             *StructureElementHeight,
                             StructureElementData);   
  } // case

  break;

  case FILTER_ERODE:
  {
   FirstPass = new CErosion(*StructureElementWidth, 
                            *StructureElementHeight,
                            StructureElementData);
  } // case

  break;

  case FILTER_CLOSE:
  {
   FirstPass = new CDilation(*StructureElementWidth, 
                             *StructureElementHeight,
                             StructureElementData);

   SecondPass = new CErosion(*StructureElementWidth, 
                             *StructureElementHeight,
                             StructureElementData);
  } // case

  break;

  case FILTER_OPEN:
  {
   FirstPass = new CErosion(*StructureElementWidth, 
                            *StructureElementHeight,
                            StructureElementData);

   SecondPass = new CDilation(*StructureElementWidth, 
                              *StructureElementHeight,
                              StructureElementData);
  } // case

  break;

  case FILTER_TOP_HAT:
  {
   FirstPass = new CErosion(*StructureElementWidth, 
                            *StructureElementHeight,
                            StructureElementData);

   SecondPass = new CDilation(*StructureElementWidth, 
                              *StructureElementHeight,
                              StructureElementData);
  } // case

  break;

  case FILTER_BOTTOM_HAT:
  {
   FirstPass = new CDilation(*StructureElementWidth, 
                             *StructureElementHeight,
                             StructureElementData);

   SecondPass = new CErosion(*StructureElementWidth, 
                             *StructureElementHeight,
                             StructureElementData);
  } // case

  break;

  default:
  {
   return -1;
  } // default

  break;
 } // switch

 // The image is filtered now. If a top hat or bottom hat filter is applied, the difference
 // between the filtered and original image is calculated and stored in the filtered image.

 if(FirstPass != NULL)
 {
  FirstPass->FilterImage(*NumberOfColumns, *NumberOfRows, OriginalImage, FilteredImage);
 } // if

 if(SecondPass != NULL)
 {
  SecondPass->FilterImage(*NumberOfColumns, *NumberOfRows, FilteredImage, FilteredImage);

  switch(*FilterType)
  {
   case FILTER_TOP_HAT:
   {
    TUInt NumberOfPixels = (*NumberOfColumns) * (*NumberOfRows);

    for(TUInt n = 0; n < NumberOfPixels; n++)
    {
     FilteredImage[n] = OriginalImage[n] - FilteredImage[n]; 
    } // for
   } // case 
   
   break;

   case FILTER_BOTTOM_HAT:
   {
    TUInt NumberOfPixels = (*NumberOfColumns) * (*NumberOfRows);

    for(TUInt n = 0; n < NumberOfPixels; n++)
    {
     FilteredImage[n] = FilteredImage[n] - OriginalImage[n];
    } // for
   } // case

   break;

   default:
   {
   } // default 

   break;
  } // switch
 } // if

 // Memory is released.

 if(FirstPass != NULL)
 {
  delete FirstPass;
 } // if

 if(SecondPass != NULL)
 {
  delete SecondPass;
 } // if

 return 0;
} // Dilate