//////////////////////////////////////////////////////////////////////////// 
// IPD - Image Processing Design Toolbox
//
// Copyright (c) by Dr. Eng. (J) Harald Galda, 2009
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
////////////////////////////////////////////////////////////////////////////


#include "IPD.h"


// SearchObjects searches connected areas in a logical image. All pixels that belong to the 
// same area get the same label. Pixels belonging to different areas get different labels.
// Background pixels get the label zero. This function is intended to be called by Scilab. 
// Calling syntax in Scilab:
// BlobImage = call('SearchObjects', ...
//                  LogicalImage, ...
//                  1, ...
//                  'd', ...
//                  NumberOfColumns, ...
//                  2, ...
//                  'i', ...
//                  NumberOfRows, ...
//                  3, ...
//                  'i', ...
//                  'out', ...
//                  [NumberOfRows, NumberOfColumns], ...
//                  4, ...
//                  'd');
extern "C" __declspec(dllexport)
int SearchObjects(double* LogicalImage,
                  TUInt* NumberOfColumns, 
                  TUInt* NumberOfRows, 
                  double* BlobImage)
{
 // initializations

 TULong NumberOfPixels = (*NumberOfRows) * (*NumberOfColumns);

 TVector ObjectImage(NumberOfPixels, 0); // pixels arranged as in Scilab

 TVector SearchTree(1, 0); // first element is zero to avoid index arithmetics problems

 TULong NextObjectLabel = 1;

 TULong PixelIndex = 0; 

 // For each true-valued pixel in LogicalImage a first label is searched. 
 for(TUInt Column = 0; Column < *NumberOfColumns; Column++)
 {
  for(TUInt Row = 0; Row < *NumberOfRows; Row++)
  {
   if(LogicalImage[PixelIndex] > 0)
   {
    // The prior (upper, upper left, left, lower left) neighbors of current pixel 
    // are checked.

    const unsigned short NumberOfCheckedNeighbors = 4;

    TULong ListOfNeighbors[4]; // list of labels at neighboring pixels

    TUShort NumberOfNeighbors = 0; // number of currently checked neighbor pixels

    // upper neighbor
    if((Row > 0) & (LogicalImage[PixelIndex - 1] > 0))
    {
     ListOfNeighbors[NumberOfNeighbors] = ObjectImage[PixelIndex - 1];

     NumberOfNeighbors++;
    } // if

    // upper left neighbor
    if((Column > 0) & (Row > 0))
    {
     TULong NeighborIndex = (Column - 1) * (*NumberOfRows) + Row - 1;

     if(LogicalImage[NeighborIndex] > 0)
     {
      ListOfNeighbors[NumberOfNeighbors] = ObjectImage[NeighborIndex];

      NumberOfNeighbors++;
     } // if
    } // if

    // left neighbor
    if(Column > 0)
    {
     TULong NeighborIndex = (Column - 1) * (*NumberOfRows) + Row;

     if(LogicalImage[NeighborIndex] > 0)
     {
      ListOfNeighbors[NumberOfNeighbors] = ObjectImage[NeighborIndex];

      NumberOfNeighbors++;
     } // if
    } // if

    // lower left neighbor
    if((Column > 0) & (Row < *NumberOfRows - 1))
    {
     TULong NeighborIndex = (Column - 1) * (*NumberOfRows) + Row + 1;

     if(LogicalImage[NeighborIndex] > 0)
     {
      ListOfNeighbors[NumberOfNeighbors] = ObjectImage[NeighborIndex];

      NumberOfNeighbors++;
     } // if
    } // if

    // If ListOfNeighbors not empty, the minimum neighbor label is selected. 
    // Otherwise the current pixel belongs to a new blob.
    if(NumberOfNeighbors > 0)
    {
     TULong CurrentLabel = ListOfNeighbors[0];

     for(TUShort n = 1; n < NumberOfNeighbors; n++)
     {
      if(ListOfNeighbors[n] < CurrentLabel)
      {
       CurrentLabel = ListOfNeighbors[n];
      } // if
     } // for

     ObjectImage[PixelIndex] = CurrentLabel;

     // All prior neighbors of the current pixel shall have the same root as the current
     // pixel.

     TULong CurrentRoot = SearchRoot(CurrentLabel, SearchTree);

     for(TUShort n = 0; n < NumberOfNeighbors; n++)
     {
      TULong NeighborRoot = SearchRoot(ListOfNeighbors[n], SearchTree);

      if(NeighborRoot != CurrentRoot)
      {
       SearchTree[NeighborRoot] = CurrentRoot;
      } // if
     } // for

     NumberOfNeighbors = 0;
    }
    else
    {
     ObjectImage[PixelIndex] = NextObjectLabel;

     SearchTree.push_back(NextObjectLabel);

     NextObjectLabel++;
    } // if
   } 
   else
   {
    BlobImage[PixelIndex] = 0;
   }// if

   // PixelIndex is incremented.
   PixelIndex++;
  } // for
 } // for

 // If any blobs were found, the adjacent blobs are merged and their labels are made 
 // continous.
 if(NextObjectLabel > 1)
 {
  // Adjacent blobs are merged.
  for(TULong PixelIndex = 0; PixelIndex < NumberOfPixels; PixelIndex++)
  {
   ObjectImage[PixelIndex] = SearchRoot(ObjectImage[PixelIndex], SearchTree);
  } // for

  // A sorted list of unique root labels is built.

  TList RootList;

  TVector::iterator TreeIndex;

  for(TreeIndex = SearchTree.begin(); TreeIndex != SearchTree.end(); TreeIndex++)
  {
   RootList.push_back(*TreeIndex);
  } // for

  RootList.sort();

  RootList.unique();

  TVector RootVector;

  TList::iterator ListIndex;

  for(ListIndex = RootList.begin(); ListIndex != RootList.end(); ListIndex++)
  {
   RootVector.push_back(*ListIndex);
  } // if

  for(TULong PixelIndex = 0; PixelIndex < NumberOfPixels; PixelIndex++)
  {
   if(LogicalImage[PixelIndex] > 0)
   {
    TULong BlobLabel = ObjectImage[PixelIndex]; 

    // Pixel is searched in RootVector.

    TULong Start = 1;

    TULong Stop = RootVector.size() - 1;

    while(Start < Stop)
    {
     TULong First = RootVector[Start];

     TULong Last = RootVector[Stop];

     if((BlobLabel != First) & (BlobLabel != Last))
     {
      TULong Mid = (Start + Stop) / 2;

      if(BlobLabel <= RootVector[Mid])
      {
       Stop = Mid;
      }
      else
      {
       Start = Mid;
      } // if
     }
     else
     {
      if(BlobLabel == First)
      {
       Stop = Start;       
      }
      else
      {
       Start = Stop;
      }// if
     } // if
    } // while

    BlobImage[PixelIndex] = Start;
   } // if
  } // for
 } // if

 return 0;
} // SearchObjects

